#include <stdio.h>
#include <hash_map>
#include <unistd.h>
#include <string>
#include <vector>
#include "logtools.h"

#define MAX_FDS  1000
#define NUM_CHILDREN  10
FILE *children[NUM_CHILDREN];
int num_fds;

using namespace std;

typedef struct
{
  string domain;
  int significant_parts;
} CFG_TYPE;

vector<CFG_TYPE> config;

struct eqstr
{
  bool operator()(string s1, string s2) const
  {
    return s1 == s2;
  }
};

struct hash_string
{
  size_t operator()(const string str) const { return hash<const char *>()(str.c_str()); }
};

typedef hash_map<const string, FILE **, hash_string, eqstr> HASH_TYPE;
HASH_TYPE fd_map;

typedef FILE * PFILE;

void usage()
{
  fprintf(stderr, "usage: clfdomainsplit [filenames]\n"
  "This program splits CLF format web logs into multiple files, one for each\n"
  "domain in the input.\n"
  "[-i input] [-d defaultfile] [-c cfg-file] [-o directory]\n");
  exit(ERR_PARAM);
}

int forkit();
bool getDomain(string &dom, PCCHAR buf);
bool printLine(string &dom, PCCHAR buf);
void closeit();
bool newdom(string &dom, CPCCHAR buf);
bool parseConfig(CPCCHAR configFile);
FILE **openOutput(string &dom);

int main(int argc, char **argv)
{
  for(int i = 0; i < NUM_CHILDREN; i++)
  {
    children[i] = NULL;
  }
  num_fds = 0;
  PCCHAR configFile = "/etc/clfdomainsplit.cfg";
  FILE *input = stdin;
  FILE *defFile = stderr;
  int int_c;
  while(-1 != (int_c = getopt(argc, argv, "c:d:i:o:")) )
  {
    switch(char(int_c))
    {
      case '?':
      case ':':
        usage();
      break;
      case 'c':
        configFile = optarg;
      break;
      case 'd':
        defFile = fopen(optarg, "a");
        if(!defFile)
        {
          fprintf(stderr, "Can't open \"%s\".\n", optarg);
          usage();
        }
      break;
      case 'i':
        input = fopen(optarg, "r");
        if(!input)
        {
          fprintf(stderr, "Can't open \"%s\".\n", optarg);
          usage();
        }
      break;
      case 'o':
        if(chdir(optarg))
        {
          fprintf(stderr, "Can't chdir() to \"%s\".\n", optarg);
          usage();
        }
      break;
    }
  }
  if(parseConfig(configFile))
    return ERR_PARAM;

  char buf[4096];
  while(fgets(buf, sizeof(buf), input))
  {
    buf[sizeof(buf) - 1] = '\0';
    strtok(buf, "\r\n");
    string dom;
    if(getDomain(dom, buf))
    {
      fprintf(stderr, "bad line:%s\n", buf);
    }
    else
    {
      if(printLine(dom, buf))
      {
        closeit();
        return 1;
      }
    }
  }
  closeit();
  return 0;
}

bool parseConfig(CPCCHAR configFile)
{
  FILE *fp = fopen(configFile, "r");
  if(!fp)
  {
    fprintf(stderr, "Can't open \"%s\".\n", configFile);
    usage();
  }
  char buf[1024];
  CFG_TYPE cfg;
  while(fgets(buf, sizeof(buf), fp))
  {
    buf[sizeof(buf) - 1] = '\0';
    strtok(buf, "\r\n");
    if(strlen(buf) && buf[0] != '#')
    {
      cfg.significant_parts = atoi(buf);
      if(cfg.significant_parts <= 0 || cfg.significant_parts > 10)
      {
        fprintf(stderr, "Error in config line \"%s\".\n", buf);
        return true;
      }
      char *ptr = index(buf, ':');
      ptr++;
      if(*ptr == '\0')
      {
        fprintf(stderr, "Error in config line \"%s\".\n", buf);
        return true;
      }
      cfg.domain = string(ptr);
      config.push_back(cfg);
    }
  }
  fclose(fp);
  return false;
}

bool printLine(string &dom, PCCHAR buf)
{
  if(!fd_map[dom])
  {
    if(newdom(dom, buf))
      return true;
  }
  fprintf(*(fd_map[dom]), "%s\n", buf);
  return false;
}

bool getDomain(string &dom, PCCHAR buf)
{
  if(!(buf = index(buf, '[')) )
    return true;
  if(!(buf = index(buf, ']')) )
    return true;
  if(!(buf = index(buf, '"')) )
    return true;
  if(!(buf = strstr(buf, "http://")) )
    return true;
  buf += 7;

  PCCHAR end = index(buf, '/');
  if(!end)
    return true;
  dom = string(buf, int(end - buf));
  return false;
}

bool newdom(string &dom, CPCCHAR buf)
{
  int significant_parts = 2;
  unsigned int i;
  for(i = 0; i < config.size(); i++)
  {
    size_t pos = dom.rfind(config[i].domain);
    if(int(pos) != -1 && (pos + config[i].domain.length()) == dom.length())
    {
      significant_parts = config[i].significant_parts;
      break;
    }
//    fprintf(*(fd_map[dom]), "%s\n", buf);
  }
  string canon("");
  for(i = dom.length(); i > 0; i--)
  {
    if(dom[i] == '.')
    {
      significant_parts--;
      if(!significant_parts)
      {
        canon.assign(dom, i + 1, dom.length() - i - 1);
        break;
      }
    }
  }
  if(fd_map[canon])
  {
    fd_map[dom] = fd_map[canon];
    return false;
  }
  FILE **fpp = openOutput(canon);
  if(!fpp)
    return true;
  fd_map[dom] = fpp;
  return false;
}

FILE **openOutput(string &dom)
{
  FILE *fp = fopen(dom.c_str(), "a");
  if(!fp)
  {
    fprintf(stderr, "Can't open/create file.\n");
    return NULL;
  }
  FILE **val = new PFILE;
  *val = fp;
  fd_map[dom] = val;
  return val;
}

int forkit(string &dom, CPCCHAR buf)
{
  int filedes[2];
//  if(pipe(filedes))
  return 0;
}

void closeit()
{
  while(!fd_map.empty())
  {
    HASH_TYPE::iterator st = fd_map.begin();
    FILE **fpp = (*st).second;
    if(fpp)
    {
      if(*fpp)
      {
        fclose(*fpp);
      }
      *fpp = NULL;
    }
    fd_map.erase(st);
  }
}
